"""Tests for doa Methods."""

import unittest
import numpy as np
from jaqpotpy.descriptors.molecular import RDKitDescriptors
from jaqpotpy.doa import Leverage, MeanVar, BoundingBox


class TestDoa(unittest.TestCase):
    def test_leverage(self):
        mols = [
            "O=C1CCCN1Cc1cccc(C(=O)N2CCC(C3CCNC3)CC2)c1",
            "O=C1CCc2cc(C(=O)N3CCC(C4CCNC4)CC3)ccc2N1",
            "CCC(=O)Nc1ccc(N(Cc2ccccc2)C(=O)n2nnc3ccccc32)cc1",
            "COc1ccc2c(N)nn(C(=O)Cc3cccc(Cl)c3)c2c1",
            "Cc1nn(C)c2[nH]nc(NC(=O)Cc3cccc(Cl)c3)c12",
            "O=C(Cc1cncc2ccccc12)N(CCC1CCCCC1)c1cccc(Cl)c1",
            "COc1ccc(N(Cc2ccccc2)C(=O)Cc2c[nH]c3ccccc23)cc1",
            "CC(C)(C)c1ccc(N(C(=O)c2ccco2)[C@H](C(=O)NCCc2cccc(F)c2)c2cccnc2)cc1",
            "OC[C@@H](O1)[C@@H](O)[C@H](O)[C@@H]2[C@@H]1c3c(O)c(OC)c(O)cc3C(=O)O2",
            "Cc1ccncc1NC(=O)Cc1cc(Cl)cc(-c2cnn(C)c2C(F)F)c1",
            "Cc1cc(C(F)(F)F)nc2c1c(N)nn2C(=O)Cc1cccc(Cl)c1",
            "Cc1cc(C(F)(F)F)nc2c1c(N)nn2C(=O)C1CCOc2ccc(Cl)cc21",
            "O=C(c1cc(=O)[nH]c2ccccc12)N1CCN(c2cccc(Cl)c2)C(=O)C1",
            "O=C1NC2(CCOc3ccc(Cl)cc32)C(=O)N1c1cncc2ccccc12",
            "COCCNC(=O)[C@@H](c1ccccc1)N1Cc2ccccc2C1=O",
            "CNCC1CCCN(C(=O)[C@@H](c2ccccc2)N2Cc3ccccc3C2=O)C1",
            "O=C1NC2(CCOc3ccc(Cl)cc32)C(=O)N1c1cncc2ccccc12",
            "COc1ccc2c(NC(=O)C3CCOc4ccc(Cl)cc43)[nH]nc2c1",
            "O=C(NC1N=Nc2ccccc21)C1CCOc2ccc(Cl)cc21",
            "COc1ccccc1OC1CCN(C(=O)c2cc(=O)[nH]c3ccccc23)C1",
            "O=C(Cc1cc(Cl)cc(Cc2ccn[nH]2)c1)Nc1cncc2ccccc12",
            "CN(C)c1ccc(N(Cc2ccsc2)C(=O)Cc2cncc3ccccc23)cc1",
            "C[C@H]1COc2ccc(Cl)cc2[C@@H]1C(=O)Nc1cncc2ccccc12",
        ]

        featurizer = RDKitDescriptors(use_fragment=False, ipc_avg=False)
        descriptors = featurizer(mols)

        doa = Leverage()
        doa.fit(descriptors)
        mol = [
            "C[C@@](C)(O1)C[C@@H](O)[C@@]1(O2)[C@@H](C)[C@@H]3CC=C4[C@]3(C2)C(=O)C[C@H]5[C@H]4CC[C@@H](C6)[C@]5(C)Cc(n7)c6nc(C[C@@]89(C))c7C[C@@H]8CC[C@@H]%10[C@@H]9C[C@@H](O)[C@@]%11(C)C%10=C[C@H](O%12)[C@]%11(O)[C@H](C)[C@]%12(O%13)[C@H](O)C[C@@]%13(C)CO",
            "COc1ccc2c(N)nn(C(=O)Cc3cccc(Cl)c3)c2c1",
            "CN(C)c1ccc(N(Cc2ccsc2)C(=O)Cc2cncc3ccccc23)cc1",
        ]

        descriptors = featurizer(mol)
        calc = doa.predict(descriptors)

        assert len(calc) == len(
            mol
        ), f"Expected len(calc) == len(mol), got {len(calc)} != {len(mol)}"
        assert (
            abs(doa.h_star - 16.434782608695652) < 0.00001
        ), f"Expected doa.a == 16.434782608695652, got {doa.h_star} != 16.434782608695652"
        assert not calc[0][
            "inDoa"
        ], f"Expected calc[0]['inDoa'] == False, got {calc[0]['inDoa']} != False"
        assert calc[1][
            "inDoa"
        ], f"Expected calc[0]['inDoa'] == True, got {calc[1]['inDoa']} != True"
        assert calc[2][
            "inDoa"
        ], f"Expected calc[0]['inDoa'] == True, got {calc[2]['inDoa']} != True"

    def test_MeanVar(self):
        mols = [
            "C[C@@](C)(O1)C[C@@H](O)[C@@]1(O2)[C@@H](C)[C@@H]3CC=C4[C@]3(C2)C(=O)C[C@H]5[C@H]4CC[C@@H](C6)[C@]5(C)Cc(n7)c6nc(C[C@@]89(C))c7C[C@@H]8CC[C@@H]%10[C@@H]9C[C@@H](O)[C@@]%11(C)C%10=C[C@H](O%12)[C@]%11(O)[C@H](C)[C@]%12(O%13)[C@H](O)C[C@@]%13(C)CO",
            "O=C1CCCN1Cc1cccc(C(=O)N2CCC(C3CCNC3)CC2)c1",
            "O=C1CCc2cc(C(=O)N3CCC(C4CCNC4)CC3)ccc2N1",
            "CCC(=O)Nc1ccc(N(Cc2ccccc2)C(=O)n2nnc3ccccc32)cc1",
            "COc1ccc2c(N)nn(C(=O)Cc3cccc(Cl)c3)c2c1",
            "Cc1nn(C)c2[nH]nc(NC(=O)Cc3cccc(Cl)c3)c12",
            "O=C(Cc1cncc2ccccc12)N(CCC1CCCCC1)c1cccc(Cl)c1",
            "COc1ccc(N(Cc2ccccc2)C(=O)Cc2c[nH]c3ccccc23)cc1",
            "CC(C)(C)c1ccc(N(C(=O)c2ccco2)[C@H](C(=O)NCCc2cccc(F)c2)c2cccnc2)cc1",
            "OC[C@@H](O1)[C@@H](O)[C@H](O)[C@@H]2[C@@H]1c3c(O)c(OC)c(O)cc3C(=O)O2",
            "Cc1ccncc1NC(=O)Cc1cc(Cl)cc(-c2cnn(C)c2C(F)F)c1",
            "Cc1cc(C(F)(F)F)nc2c1c(N)nn2C(=O)Cc1cccc(Cl)c1",
            "Cc1cc(C(F)(F)F)nc2c1c(N)nn2C(=O)C1CCOc2ccc(Cl)cc21",
            "O=C(c1cc(=O)[nH]c2ccccc12)N1CCN(c2cccc(Cl)c2)C(=O)C1",
            "O=C1NC2(CCOc3ccc(Cl)cc32)C(=O)N1c1cncc2ccccc12",
        ]

        featurizer = RDKitDescriptors(use_fragment=False, ipc_avg=False)
        descriptors = featurizer(mols)

        doa = MeanVar()
        doa.fit(descriptors)

        mol = [
            "C[C@@](C)(O1)C[C@@H](O)[C@@]1(O2)[C@@H](C)[C@@H]3CC=C4[C@]3(C2)C(=O)C[C@H]5[C@H]4CC[C@@H](C6)[C@]5(C)Cc(n7)c6nc(C[C@@]89(C))c7C[C@@H]8CC[C@@H]%10[C@@H]9C[C@@H](O)[C@@]%11(C)C%10=C[C@H](O%12)[C@]%11(O)[C@H](C)[C@]%12(O%13)[C@H](O)C[C@@]%13(C)CO",
            "CCC",
        ]
        descriptors = featurizer(mol)
        calc = doa.predict(descriptors)
        diag = np.diag(doa.bounds)

        assert len(calc) == len(mol)
        assert not calc[0][
            "inDoa"
        ], f"Expected calc[0]['inDoa'] == False, got {calc[0]['inDoa']} != False"
        assert not calc[1][
            "inDoa"
        ], f"Expected calc[0]['inDoa'] == False, got {calc[1]['inDoa']} != False"
        assert np.allclose(
            diag, [1.31511044e01, 6.69162726e-01, 5.37187947e-03], atol=1e-5
        ), f"Expected diag == [1.31511044e+01, 6.69162726e-01, 5.37187947e-03], got diag != {diag}"

    def test_BoundingBox(self):
        mols = [
            "C[C@@](C)(O1)C[C@@H](O)[C@@]1(O2)[C@@H](C)[C@@H]3CC=C4[C@]3(C2)C(=O)C[C@H]5[C@H]4CC[C@@H](C6)[C@]5(C)Cc(n7)c6nc(C[C@@]89(C))c7C[C@@H]8CC[C@@H]%10[C@@H]9C[C@@H](O)[C@@]%11(C)C%10=C[C@H](O%12)[C@]%11(O)[C@H](C)[C@]%12(O%13)[C@H](O)C[C@@]%13(C)CO",
            "O=C1CCCN1Cc1cccc(C(=O)N2CCC(C3CCNC3)CC2)c1",
            "O=C1CCc2cc(C(=O)N3CCC(C4CCNC4)CC3)ccc2N1",
            "CCC(=O)Nc1ccc(N(Cc2ccccc2)C(=O)n2nnc3ccccc32)cc1",
            "COc1ccc2c(N)nn(C(=O)Cc3cccc(Cl)c3)c2c1",
            "Cc1nn(C)c2[nH]nc(NC(=O)Cc3cccc(Cl)c3)c12",
            "O=C(Cc1cncc2ccccc12)N(CCC1CCCCC1)c1cccc(Cl)c1",
            "COc1ccc(N(Cc2ccccc2)C(=O)Cc2c[nH]c3ccccc23)cc1",
            "CC(C)(C)c1ccc(N(C(=O)c2ccco2)[C@H](C(=O)NCCc2cccc(F)c2)c2cccnc2)cc1",
            "OC[C@@H](O1)[C@@H](O)[C@H](O)[C@@H]2[C@@H]1c3c(O)c(OC)c(O)cc3C(=O)O2",
            "Cc1ccncc1NC(=O)Cc1cc(Cl)cc(-c2cnn(C)c2C(F)F)c1",
            "Cc1cc(C(F)(F)F)nc2c1c(N)nn2C(=O)Cc1cccc(Cl)c1",
            "Cc1cc(C(F)(F)F)nc2c1c(N)nn2C(=O)C1CCOc2ccc(Cl)cc21",
            "O=C(c1cc(=O)[nH]c2ccccc12)N1CCN(c2cccc(Cl)c2)C(=O)C1",
            "O=C1NC2(CCOc3ccc(Cl)cc32)C(=O)N1c1cncc2ccccc12",
        ]

        featurizer = RDKitDescriptors(use_fragment=False, ipc_avg=False)
        descriptors = featurizer(mols)
        print(type(descriptors))
        doa = BoundingBox()
        doa.fit(descriptors)

        mol = [
            "C[C@@](C)(O1)C[C@@H](O)[C@@]1(O2)[C@@H](C)[C@@H]3CC=C4[C@]3(C2)C(=O)C[C@H]5[C@H]4CC[C@@H](C6)[C@]5(C)Cc(n7)c6nc(C[C@@]89(C))c7C[C@@H]8CC[C@@H]%10[C@@H]9C[C@@H](O)[C@@]%11(C)C%10=C[C@H](O%12)[C@]%11(O)[C@H](C)[C@]%12(O%13)[C@H](O)C[C@@]%13(C)CO",
            "CCC",
        ]
        descriptors = featurizer(mol)
        calc = doa.predict(descriptors)
        first_feature_bounds = doa.bounding_box[0]
        last_feature_bounds = doa.bounding_box[-1]
        print("Bounding box")
        assert len(calc) == len(mol)
        assert calc[0][
            "inDoa"
        ], f"Expected calc[0]['inDoa'] == True, got {calc[0]['inDoa']} != True"
        assert not calc[1][
            "inDoa"
        ], f"Expected calc[0]['inDoa'] == False, got {calc[1]['inDoa']} != False"
        assert np.allclose(
            first_feature_bounds, [12.0648171, 14.92728396], atol=1e-5
        ), f"Expected first_feature_bounds == [12.0648171 , 14.92728396], got {first_feature_bounds} != [12.0648171 , 14.92728396]"
        assert np.allclose(
            last_feature_bounds, [7.22405000e01, 2.40083000e02], atol=1e-5
        ), f"Expected last_feature_bounds == [7.22405000e+01,  2.40083000e+02], got {last_feature_bounds} !=  [7.22405000e+01,  2.40083000e+02]"



# flake8: noqa

# import apis into api package
from jaqpotpy.api.openapi.api.api_keys_api import ApiKeysApi
from jaqpotpy.api.openapi.api.auth_api import AuthApi
from jaqpotpy.api.openapi.api.dataset_api import DatasetApi
from jaqpotpy.api.openapi.api.feature_api import FeatureApi
from jaqpotpy.api.openapi.api.lead_api import LeadApi
from jaqpotpy.api.openapi.api.model_api import ModelApi
from jaqpotpy.api.openapi.api.organization_api import OrganizationApi
from jaqpotpy.api.openapi.api.organization_invitation_api import OrganizationInvitationApi




import numpy as np
import unittest
from rdkit import Chem
from rdkit.Chem import AllChem
from mordred import Chi, ABCIndex

from jaqpotpy.descriptors.molecular import MordredDescriptors


class TestMordredDescriptors(unittest.TestCase):
    """Test MordredDescriptors."""

    def setUp(self):
        """Set up tests."""
        self.smiles1 = "CC(=O)OC1=CC=CC=C1C(=O)O"
        self.mol = Chem.MolFromSmiles(self.smiles1)
        self.smiles2 = "CCCC(=O)OC1=CC=CC=C1C(=O)O"
        self.mol2 = Chem.MolFromSmiles(self.smiles2)
        self.featurizer = MordredDescriptors()

    def test_mordred_descriptors(self):
        """Test featurize using mols"""
        descriptors = self.featurizer([self.mol, self.mol2])
        assert descriptors.shape == (2, 1613), "Wrong shape"
        assert isinstance(
            descriptors[0][1], (int, float, np.number)
        ), "The value is not numeric"

    def test_mordred_descriptors_dataframe(self):
        """Test featurize_dataframe using mols"""
        descriptors = self.featurizer.featurize_dataframe([self.mol, self.mol2])
        assert descriptors.shape == (2, 1613)
        assert isinstance(descriptors.iloc[0, 1], (int, float, np.number))

    def test_mordred_descriptors_with_smiles(self):
        """Test featurize using Smiles"""
        descriptors = self.featurizer([self.smiles1, self.smiles2])
        assert descriptors.shape == (2, 1613)
        assert isinstance(descriptors[0][1], (int, float, np.number))

    def test_mordred_descriptors_with_3D_info(self):
        """Test simple descriptors with 3D info"""
        featurizer = MordredDescriptors(ignore_3D=False)
        descriptors = featurizer([self.mol, self.mol2])
        assert descriptors.shape == (2, 1826)
        assert isinstance(
            descriptors[0][1], (int, float, np.number)
        ), "The value is not numeric"

        # calculate coordinates
        mol = self.mol
        mol_with_conf = Chem.AddHs(mol)
        AllChem.EmbedMolecule(mol_with_conf, AllChem.ETKDG())
        descriptors = featurizer([mol_with_conf])
        assert descriptors.shape == (1, 1826)
        # not zero values
        assert not np.allclose(descriptors[0][780:784], np.array([0.0, 0.0, 0.0, 0.0]))

    def test_base_mordred(self):
        """Mordredcommunity is used instead of morded cause of numpy errors"""
        benzene = Chem.MolFromSmiles("c1ccccc1")
        # create descriptor instance
        abci = ABCIndex.ABCIndex()

        # calculate descriptor value
        result = abci(benzene)

        assert result == 4.242640687119286
        # create descriptor instance with parameter
        chi_pc4 = Chi.Chi(type="path_cluster", order=4)

        # calculate
        result = chi_pc4(benzene)

        assert result == 0.0



import torch
import torchvision
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

n_epochs = 3
batch_size_train = 64
batch_size_test = 1000
learning_rate = 0.01
momentum = 0.5
log_interval = 10

random_seed = 1
torch.backends.cudnn.enabled = False
torch.manual_seed(random_seed)

train_loader = torch.utils.data.DataLoader(
    torchvision.datasets.MNIST(
        "./files/",
        train=True,
        download=True,
        transform=torchvision.transforms.Compose(
            [
                torchvision.transforms.ToTensor(),
                torchvision.transforms.Normalize((0.1307,), (0.3081,)),
            ]
        ),
    ),
    batch_size=batch_size_train,
    shuffle=True,
)

test_loader = torch.utils.data.DataLoader(
    torchvision.datasets.MNIST(
        "./files/",
        train=False,
        download=True,
        transform=torchvision.transforms.Compose(
            [
                torchvision.transforms.ToTensor(),
                torchvision.transforms.Normalize((0.1307,), (0.3081,)),
            ]
        ),
    ),
    batch_size=batch_size_test,
    shuffle=True,
)


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x)


network = Net()
optimizer = optim.SGD(network.parameters(), lr=learning_rate, momentum=momentum)

train_losses = []
train_counter = []
test_losses = []
test_counter = [i * len(train_loader.dataset) for i in range(n_epochs + 1)]


def train(epoch):
    network.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        optimizer.zero_grad()
        output = network(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if batch_idx % log_interval == 0:
            print(
                "Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}".format(
                    epoch,
                    batch_idx * len(data),
                    len(train_loader.dataset),
                    100.0 * batch_idx / len(train_loader),
                    loss.item(),
                )
            )
            train_losses.append(loss.item())
            train_counter.append(
                (batch_idx * 64) + ((epoch - 1) * len(train_loader.dataset))
            )
            torch.save(network.state_dict(), "./results/model.pth")
            torch.save(optimizer.state_dict(), "./results/optimizer.pth")


def test():
    network.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            output = network(data)
            test_loss += F.nll_loss(output, target, size_average=False).item()
            pred = output.data.max(1, keepdim=True)[1]
            correct += pred.eq(target.data.view_as(pred)).sum()
    test_loss /= len(test_loader.dataset)
    test_losses.append(test_loss)
    print(
        "\nTest set: Avg. loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n".format(
            test_loss,
            correct,
            len(test_loader.dataset),
            100.0 * correct / len(test_loader.dataset),
        )
    )


test()
for epoch in range(1, n_epochs + 1):
    train(epoch)
    test()



"""Topological fingerprints."""

from typing import Dict
import pandas as pd
import numpy as np
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer


class TopologicalFingerprint(MolecularFeaturizer):
    """Topological (Morgan) fingerprints.

    Extended Connectivity Circular Fingerprints compute a bag-of-words style
    representation of a molecule by breaking it into local neighborhoods and
    hashing into a bit vector of the specified size. It is used specifically
    for structure-activity modeling. See [1]_ for more details.

    References
    ----------
    .. [1] Rogers, David, and Mathew Hahn. "Extended-connectivity fingerprints."
       Journal of chemical information and modeling 50.5 (2010): 742-754.

    Note
    ----
    This class requires RDKit to be installed.

    Examples
    --------
    >>> import jaqpotpy as jt
    >>> from rdkit import Chem
    >>> smiles = ['C1=CC=CC=C1']
    >>> # Example 1: (size = 2048, radius = 4)
    >>> featurizer = jt.descriptors.TopologicalFingerprint(size=2048, radius=4)
    >>> features = featurizer.featurize(smiles)
    """

    @property
    def __name__(self):
        return "TopologicalFingerprint"

    def __init__(
        self,
        radius: int = 2,
        size: int = 2048,
        chiral: bool = False,
        bonds: bool = True,
        features: bool = False,
    ):
        """Initialize the TopologicalFingerprint.

        Parameters
        ----------
        radius : int, optional (default=2)
            Fingerprint radius.
        size : int, optional (default=2048)
            Length of generated bit vector.
        chiral : bool, optional (default=False)
            Whether to consider chirality in fingerprint generation.
        bonds : bool, optional (default=True)
            Whether to consider bond order in fingerprint generation.
        features : bool, optional (default=False)
            Whether to use feature information instead of atom information; see
            RDKit docs for more info.
        """
        self.radius = radius
        self.size = size
        self.chiral = chiral
        self.bonds = bonds
        self.features = features
        self.col_names = None

    def _featurize(self, datapoint, convert_nan: bool = True, **kwargs) -> np.ndarray:
        """Calculate circular fingerprint.

        Parameters
        ----------
        datapoint : rdkit.Chem.rdchem.Mol
            The molecule to featurize.
        convert_nan : bool, optional (default=True)
            Whether to convert NaN values.

        Returns
        -------
        np.ndarray
            A numpy array of the circular fingerprint.
        """
        try:
            from rdkit import Chem
            from rdkit.Chem import rdMolDescriptors
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")

        fp = rdMolDescriptors.GetMorganFingerprintAsBitVect(
            datapoint,
            self.radius,
            nBits=self.size,
            useChirality=self.chiral,
            useBondTypes=self.bonds,
            useFeatures=self.features,
        )
        fp = np.asarray(fp, dtype=float)
        return fp

    def __hash__(self):
        return hash(
            (
                self.radius,
                self.size,
                self.chiral,
                self.bonds,
                self.features,
            )
        )

    def __eq__(self, other):
        if not isinstance(self, other.__class__):
            return False
        return (
            self.radius == other.radius
            and self.size == other.size
            and self.chiral == other.chiral
            and self.bonds == other.bonds
            and self.features == other.features
        )

    def featurize_dataframe(
        self, datapoints, convert_nan=False, log_every_n=1000, **kwargs
    ) -> pd.DataFrame:
        """Featurize a list of molecules and return a DataFrame.

        Parameters
        ----------
        datapoints : list
            List of molecules to featurize.
        convert_nan : bool, optional (default=False)
            Whether to convert NaN values.
        log_every_n : int, optional (default=1000)
            Log progress every n molecules.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the featurized molecules.
        """
        features = self.featurize(
            datapoints, convert_nan=True, log_every_n=1000, **kwargs
        )

        self.col_names = [f"Bit_{i}" for i in range(self.size)]
        df = pd.DataFrame(features, columns=self.col_names)
        return df



from jaqpotpy.models.torch_geometric_models.graph_neural_network import (
    GraphConvolutionNetwork,
    GraphSageNetwork,
    GraphAttentionNetwork,
    GraphTransformerNetwork,
    pyg_to_onnx,
    # pyg_to_torchscript,
)



from base64 import b64encode


def model_to_b64encoding(raw_model):
    """
    Encodes a raw model to a base64 string.

    Args:
        raw_model (bytes): The raw model data in bytes.

    Returns:
        str: The base64 encoded string of the model.
    """
    raw = b64encode(raw_model)
    model_b64encoded = raw.decode("utf-8")
    return model_b64encoded


def file_to_b64encoding(filePath):
    """
    Reads a file and encodes its contents to a base64 string.

    Args:
        filePath (str): The path to the file to be encoded.

    Returns:
        str: The base64 encoded string of the file contents.
    """
    with open(filePath, "rb") as file:
        file_data = file.read()
        b64_encoded_data = b64encode(file_data)
        return b64_encoded_data.decode("utf-8")



from jaqpotpy.api.openapi import Configuration, ApiClient


class JaqpotApiHttpClient(ApiClient):
    """
    A client for interacting with the Jaqpot API over HTTP.
    """

    def __init__(self, host=None, access_token=None):
        """
        Initialize the JaqpotApiHttpClient.

        :param host: str, optional
            The base URL of the Jaqpot API. Defaults to "https://api.jaqpot.com".
        :param access_token: str, optional
            The access token for authenticating with the Jaqpot API.
        """
        self.host = host or "https://api.jaqpot.com"
        configuration = Configuration(
            host=self.host, access_token=access_token if access_token else None
        )
        super().__init__(configuration=configuration)

    def set_access_token(self, access_token):
        """
        Set the access token for the API client.

        :param access_token: str
            The new access token.
        """
        self.configuration.access_token = access_token



import numpy as np
import inspect
import logging
from typing import Any, List, Iterable
import pandas as pd
from tqdm import tqdm


logger = logging.getLogger(__name__)
_print_threshold = 1000
verbose = False


class AbstractFeaturizer(object):
    """Abstract class for calculating a set of features for a datapoint.
    This class is abstract.
    """

    def featurize(
        self, datapoints: Iterable[Any], log_every_n: int = 1000, **kwargs
    ) -> np.ndarray:
        """Calculate features for datapoints.

        Parameters
        ----------
        datapoints: Iterable[Any]
          A sequence of objects that you'd like to featurize. Subclassses of
          `Featurizer` should instantiate the `_featurize` method that featurizes
          objects in the sequence.
        log_every_n: int, default 1000
          Logs featurization progress every `log_every_n` steps.

        Returns
        -------
        np.ndarray
          A numpy array containing a featurized representation of `datapoints`.

        """
        datapoints = list(datapoints)
        features = []
        if verbose is False:
            disable_tq = True
        else:
            disable_tq = False
        for i, point in enumerate(
            tqdm(datapoints, desc="Creating descriptors", disable=disable_tq)
        ):
            try:
                features.append(self._featurize(point, **kwargs))
            except:
                if verbose is True:
                    logger.warning(
                        "Failed to featurize datapoint %d. Appending array with nan"
                    )
                features.append(np.array([]))

        return np.asarray(features)

    def __call__(self, datapoints: Iterable[Any], **kwargs):
        """Calculate features for datapoints.
        `**kwargs` will get passed directly to `Featurizer.featurize`

        Parameters
        ----------
        datapoints: Iterable[Any]
          Any blob of data you like. Subclasss should instantiate this.

        """
        return self.featurize(datapoints, **kwargs)

    def _featurize(self, datapoint: Any, **kwargs):
        """Calculate features for a single datapoint.

        Parameters
        ----------
        datapoint: Any
          Any blob of data you like. Subclass should instantiate this.

        """
        raise NotImplementedError("Featurizer is not defined.")

    def _featurize_dataframe(self, datapoint: Any, **kwargs):
        """Calculate features for a single datapoint.

        Parameters
        ----------
        datapoint: Any
          Any blob of data you like. Subclass should instantiate this.

        """
        raise NotImplementedError("Featurizer is not defined.")

    def __repr__(self) -> str:
        """Convert self to repr representation.

        Returns
        -------
        str
          The string represents the class.

        Examples
        --------
        >>> import jaqpotpy as jt
        >>> jt.descriptors.CircularFingerprint(size=1024, radius=4)
        CircularFingerprint[radius=4, size=1024, chiral=False, bonds=True, features=False, sparse=False, smiles=False]
        >>> jt.descriptors.CGCNNFeaturizer()
        CGCNNFeaturizer[radius=8.0, max_neighbors=12, step=0.2]

        """

        args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        args_names = [arg for arg in args_spec.args if arg != "self"]
        args_info = ""
        for arg_name in args_names:
            value = self.__dict__[arg_name]
            # for str
            if isinstance(value, str):
                value = "'" + value + "'"
            # for list
            if isinstance(value, list):
                threshold = get_print_threshold()
                value = np.array2string(np.array(value), threshold=threshold)
            args_info += arg_name + "=" + str(value) + ", "
        return self.__class__.__name__ + "[" + args_info[:-2] + "]"

    def __str__(self) -> str:
        """Convert self to str representation.

        Returns
        -------
        str
          The string represents the class.
        """
        args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        args_names = [arg for arg in args_spec.args if arg != "self"]
        args_num = len(args_names)
        args_default_values = [None for _ in range(args_num)]
        if args_spec.defaults is not None:
            defaults = list(args_spec.defaults)
            args_default_values[-len(defaults) :] = defaults

        override_args_info = ""
        for arg_name, default in zip(args_names, args_default_values):
            if arg_name in self.__dict__:
                arg_value = self.__dict__[arg_name]
                # validation
                # skip list
                if isinstance(arg_value, list):
                    continue
                if isinstance(arg_value, str):
                    # skip path string
                    if "\\/." in arg_value or "/" in arg_value or "." in arg_value:
                        continue
                # main logic
                if default != arg_value:
                    override_args_info += "_" + arg_name + "_" + str(arg_value)
        return self.__class__.__name__ + override_args_info


class MolecularFeaturizer(AbstractFeaturizer):
    """Abstract class for calculating a set of features for a
    molecule. Τhε MolecularFeaturizer uses SMILES strings to represent
    small molecules. All other featurizers which are subclasses of
    this class should plan to process input which comes as smiles
    strings. Child classes need to implement the _featurize method for
    calculating features for a single molecule.

    .. [1] Ramsundar-et-al-2019,. "Deep Learning for the Life Sciences."
       O'Reilly Media (2019):.
    .. [2] https://github.com/deepchem/deepchem

    Note:
    ----
    The subclasses of this class require RDKit to be installed.

    """

    def featurize(
        self, datapoints, convert_nan: bool = True, log_every_n=1000, **kwargs
    ) -> np.ndarray:
        """Calculate features for molecules.
        Parameters: datapoints--> SMILES string
        Returns:features--> np.ndarray
        """
        try:
            from rdkit import Chem
            from rdkit.Chem import rdmolfiles
            from rdkit.Chem import rdmolops
            from rdkit.Chem.rdchem import Mol
            from math import isnan
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")

        # Special case handling of single molecule
        if isinstance(datapoints, str) or isinstance(datapoints, Mol):
            datapoints = [datapoints]
        else:
            # Convert iterables to list
            datapoints = list(datapoints)

        features: list = []
        if verbose is False:
            disable_tq = True
        else:
            disable_tq = False
        for i, mol in enumerate(
            tqdm(datapoints, desc="Creating descriptors", disable=disable_tq)
        ):
            if i % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % i)

            try:
                if isinstance(mol, str):
                    # mol must be a RDKit Mol object, so parse a SMILES
                    mol = Chem.MolFromSmiles(mol)
                    # SMILES is unique, so set a canonical order of atoms
                    new_order = rdmolfiles.CanonicalRankAtoms(mol)
                    mol = rdmolops.RenumberAtoms(mol, new_order)

                features.append(self._featurize(mol, **kwargs))
            except Exception as e:
                if isinstance(mol, Chem.rdchem.Mol):
                    mol = Chem.MolToSmiles(mol)
                if verbose is True:
                    logger.warning(
                        "Failed to featurize datapoint %d, %s. Appending array with nan values",
                        i,
                        mol,
                    )
                    logger.warning("Exception message: {}".format(e))
                features.append(np.array([]))

        max_descriptor_len = max([len(f) for f in features])
        for i, feature in enumerate(features):
            if len(feature) == 0:
                features[i] = np.full(max_descriptor_len, np.nan)
        return np.asarray(features)

    def featurize_dataframe(
        self, datapoints, convert_nan: bool = True, log_every_n=1000, **kwargs
    ) -> Any:
        """Calculate features for molecules.

        Parameters
        ----------
        datapoints:  SMILES string
        log_every_n: int, default 1000
          Logging messages reported every `log_every_n` samples.

        Returns
        -------
        features: pd.Dataframe()
          A pandas Dataframe containing a featurized representation of `datapoints`.

        """

        raise NotImplementedError("Featurizer is not defined.")


def get_print_threshold() -> int:
    """Return the printing threshold for datasets.
    The print threshold is the number of elements from ids/tasks to
    print when printing representations of `Dataset` objects.

    Returns
    -------
    threshold: int
      Number of elements that will be printed

    """
    return _print_threshold



import webbrowser

from keycloak import KeycloakOpenID

import jaqpotpy
from jaqpotpy.api.get_installed_libraries import get_installed_libraries
from jaqpotpy.api.jaqpot_api_client_builder import JaqpotApiHttpClientBuilder
from jaqpotpy.api.jaqpot_api_http_client import JaqpotApiHttpClient
from jaqpotpy.api.model_to_b64encoding import model_to_b64encoding
from jaqpotpy.api.openapi.api.model_api import ModelApi
from jaqpotpy.api.openapi.models.feature import Feature
from jaqpotpy.api.openapi.models.feature_type import FeatureType
from jaqpotpy.api.openapi.models.model import Model
from jaqpotpy.api.openapi.models.model_task import ModelTask
from jaqpotpy.api.openapi.models.model_type import ModelType
from jaqpotpy.api.openapi.models.model_visibility import ModelVisibility
from jaqpotpy.helpers.logging import init_logger
from jaqpotpy.helpers.url_utils import add_subdomain

ENCODING = "utf-8"


class Jaqpot:
    """Deploys sklearn and PyTorch models on Jaqpot.

    This class provides methods to log in to Jaqpot using Keycloak and deploy
    machine learning models (sklearn and PyTorch) on the Jaqpot platform.

    Parameters
    ----------
    base_url : str, optional
        The base URL on which Jaqpot services are deployed. Default is "https://jaqpot.org".
    app_url : str, optional
        The URL for the Jaqpot application. If not provided, it is derived from the base URL.
    login_url : str, optional
        The URL for the Jaqpot login. If not provided, it is derived from the base URL.
    api_url : str, optional
        The URL for the Jaqpot API. If not provided, it is derived from the base URL.
    keycloak_realm : str, optional
        The Keycloak realm name. Default is "jaqpot".
    keycloak_client_id : str, optional
        The Keycloak client ID. Default is "jaqpot-client".
    create_logs : bool, optional
        Whether to create logs. Default is False.
    """

    def __init__(
        self,
        base_url=None,
        app_url=None,
        login_url=None,
        api_url=None,
        keycloak_realm=None,
        keycloak_client_id=None,
        create_logs=False,
    ):
        # logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)
        self.log = init_logger(
            __name__, testing_mode=False, output_log_file=create_logs
        )
        if base_url:
            self.base_url = base_url
        else:
            self.base_url = "https://jaqpot.org"
        self.app_url = app_url or add_subdomain(self.base_url, "app")
        self.login_url = login_url or add_subdomain(self.base_url, "login")
        self.api_url = api_url or add_subdomain(self.base_url, "api")
        self.keycloak_realm = keycloak_realm or "jaqpot"
        self.keycloak_client_id = keycloak_client_id or "jaqpot-client"
        self.access_token = None
        self.http_client = None

    def login(self):
        """
        Log in to Jaqpot using Keycloak.

        This method opens a browser window for the user to log in via Keycloak,
        then exchanges the authorization code for an access token.

        Returns
        -------
        None
        """
        # Configure Keycloak client
        keycloak_openid = KeycloakOpenID(
            server_url=self.login_url,
            client_id=self.keycloak_client_id,
            realm_name=self.keycloak_realm,
        )

        # Generate the authorization URL
        redirect_uri = "urn:ietf:wg:oauth:2.0:oob"
        auth_url = keycloak_openid.auth_url(
            redirect_uri=redirect_uri,
            scope="openid email profile",
            state="random_state_value",
        )

        print(f"Open this URL in your browser and log in:\n{auth_url}")

        # Automatically open the browser (optional)
        webbrowser.open(auth_url)

        code = input("Enter the authorization code you received: ")

        # Exchange the code for an access token
        token_response = keycloak_openid.token(
            grant_type="authorization_code", code=code, redirect_uri=redirect_uri
        )

        access_token = token_response["access_token"]
        self.access_token = access_token
        self.http_client = (
            JaqpotApiHttpClientBuilder(host=self.api_url)
            .build_with_access_token(self.access_token)
            .build()
        )

    def deploy_sklearn_model(self, model, name, description, visibility):
        """
        Deploy an sklearn model on Jaqpot.

        Parameters
        ----------
        model : object
            The sklearn model to be deployed. The model should have attributes
            like `onnx_model`, `onnx_preprocessor`, `type`, `jaqpotpy_version`,
            `doa_data`, `libraries`, `dependentFeatures`, `independentFeatures`,
            `selected_features`, `featurizers`, `preprocessors`, and `scores`.
        name : str
            The name of the model.
        description : str
            A description of the model.
        visibility : str
            The visibility of the model (e.g., 'public', 'private').

        Returns
        -------
        None
        """
        model_api = ModelApi(self.http_client)
        raw_model = model_to_b64encoding(model.onnx_model.SerializeToString())
        raw_preprocessor = (
            model_to_b64encoding(model.onnx_preprocessor.SerializeToString())
            if model.onnx_preprocessor
            else None
        )
        body_model = Model(
            name=name,
            type=model.type,
            jaqpotpy_version=model.jaqpotpy_version,
            doas=model.doa_data,
            libraries=model.libraries,
            dependent_features=[
                Feature(
                    key=feature_i["key"],
                    name=feature_i["name"],
                    feature_type=feature_i["featureType"],
                )
                for feature_i in model.dependentFeatures
            ],
            independent_features=[
                Feature(
                    key=feature_i["key"],
                    name=feature_i["name"],
                    feature_type=feature_i["featureType"],
                    possible_values=feature_i["possible_values"]
                    if "possible_values" in feature_i
                    else None,
                )
                for feature_i in model.independentFeatures
            ],
            visibility=ModelVisibility(visibility),
            task=ModelTask(model.task.upper()),
            raw_preprocessor=raw_preprocessor,
            raw_model=raw_model,
            selected_features=model.selected_features,
            description=description,
            featurizers=model.featurizers,
            preprocessors=model.preprocessors,
            scores=model.scores,
        )
        response = model_api.create_model_with_http_info(model=body_model)
        if response.status_code < 300:
            model_url = response.headers.get("Location")
            model_id = model_url.split("/")[-1]

            self.log.info(
                "Model has been successfully uploaded. The url of the model is %s",
                self.app_url + "/dashboard/models/" + model_id,
            )
        else:
            # error = response.headers.get("error")
            # error_description = response.headers.get("error_description")
            self.log.error("Error code: " + str(response.status_code.value))

    def deploy_torch_model(
        self,
        onnx_model,
        featurizer,
        name,
        description,
        target_name,
        visibility,
        task,
    ):
        """
        Deploy a PyTorch model on Jaqpot.

        Parameters
        ----------
        onnx_model : object
            The ONNX model to be deployed.
        featurizer : object
            The featurizer used for preprocessing. The featurizer should have a method `get_dict()`.
        name : str
            The name of the model.
        description : str
            A description of the model.
        target_name : str
            The name of the target feature.
        visibility : str
            The visibility of the model (e.g., 'public', 'private').
        task : str
            The task type (e.g., 'binary_classification', 'regression', 'multiclass_classification').

        Returns
        -------
        None
        """
        if task == "binary_classification":
            model_task = ModelTask.BINARY_CLASSIFICATION
            feature_type = FeatureType.INTEGER
        elif task == "regression":
            model_task = ModelTask.REGRESSION
            feature_type = FeatureType.FLOAT
        elif task == "multiclass_classification":
            model_task = ModelTask.MULTICLASS_CLASSIFICATION
            feature_type = FeatureType.INTEGER
        else:
            raise ValueError("Task should be either classification or regression")
        model_api = ModelApi(self.http_client)
        # Change Base URL when not in local testing
        # baseurl: "http://localhost.jaqpot.org:8080/"
        torch_config = featurizer.get_dict()
        body_model = Model(
            name=name,
            type=ModelType.TORCH_ONNX,
            jaqpotpy_version=jaqpotpy.__version__,
            libraries=get_installed_libraries(),
            dependent_features=[
                Feature(key=target_name, name=target_name, feature_type=feature_type)
            ],  # TODO: Spaces dont work in endpoint name
            independent_features=[
                Feature(key="SMILES", name="SMILES", feature_type=FeatureType.SMILES)
            ],
            torch_config=torch_config,
            task=model_task,
            visibility=ModelVisibility(visibility),
            raw_model=onnx_model,
            description=description,
        )
        response = model_api.create_model_with_http_info(model=body_model)
        if response.status_code < 300:
            model_url = response.headers.get("Location")
            model_id = model_url.split("/")[-1]

            self.log.info(
                "Model has been successfully uploaded. The url of the model is %s",
                self.app_url + "/dashboard/models/" + model_id,
            )
        else:
            self.log.error("Error code: " + str(response.status_code.value))
            self.log.error("Error message: " + response.content.decode("utf-8"))



from typing import Callable, List, Optional, Any
import numpy as np
import pandas as pd
from jaqpotpy.descriptors.base_classes import MolecularFeaturizer
# warnings.filterwarnings("ignore", category=DeprecationWarning)


class MordredDescriptors(MolecularFeaturizer):
    """Mordred descriptors.

    Attributes
    ----------
    descriptors: List[str]
        List of Mordred descriptor names used in this class.

    References
    ----------
    .. [1] Moriwaki, Hirotomo, et al. "Mordred: a molecular descriptor calculator."
       Journal of cheminformatics 10.1 (2018): 4.
    .. [2] http://mordred-descriptor.github.io/documentation/master/descriptors.html

    Examples
    --------
    >>> import jaqpotpy as jt
    >>> smiles = ['CC(=O)OC1=CC=CC=C1C(=O)O']
    >>> featurizer = jt.descriptors.MordredDescriptors()
    >>> features = featurizer.featurize(smiles)
    """

    @property
    def __name__(self):
        return "MordredDescriptors"

    def __init__(self, ignore_3D: bool = True):
        """Initialize the MordredDescriptors.

        Parameters
        ----------
        ignore_3D: bool, optional (default True)
            Whether to use 3D information or not.
        """
        self.col_names = None
        self.ignore_3D = ignore_3D
        self.calc: Optional[Callable] = None
        self.descriptors: Optional[List] = None

    def _featurize(self, datapoint, convert_nan: bool = True, **kwargs) -> np.ndarray:
        """Calculate Mordred descriptors for a single molecule.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
            The molecule for which to calculate descriptors.
        convert_nan: bool, optional (default True)
            Whether to convert NaN values to -1000.0.

        Returns
        -------
        np.ndarray
            1D array of Mordred descriptors for `datapoint`.
            If ignore_3D is True, the length is 1613.
            If ignore_3D is False, the length is 1826.
        """
        if self.calc is None:
            try:
                from mordred import Calculator, descriptors, is_missing

                self.is_missing = is_missing
                self.calc = Calculator(descriptors, ignore_3D=self.ignore_3D)
                self.descriptors = list(descriptors.__all__)
            except ModuleNotFoundError:
                raise ImportError("This class requires Mordred to be installed.")

        feature = self.calc(datapoint)
        self.col_names = [key for key in feature.keys()]
        # convert errors to zero
        if convert_nan:
            feature = [
                -1000.0 if self.is_missing(val) or isinstance(val, str) else val
                for val in feature
            ]
        return np.asarray(feature)

    def featurize_dataframe(
        self, datapoints, convert_nan: bool = True, log_every_n=1000, **kwargs
    ) -> pd.DataFrame:
        """Calculate Mordred descriptors for a list of SMILES strings.

        Parameters
        ----------
        datapoints: list of str
            List of SMILES strings representing the molecules.
        convert_nan: bool, optional (default True)
            Whether to convert NaN values to -1000.0.
        log_every_n: int, optional (default 1000)
            Log progress every n molecules.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the Mordred descriptors for each molecule.
            If ignore_3D is True, the length is 1613.
            If ignore_3D is False, the length is 1826.
        """
        features = self.featurize(datapoints, convert_nan, log_every_n, **kwargs)
        df = pd.DataFrame(features, columns=self.col_names)
        return df



"""Wrappers around the logging module."""

from __future__ import absolute_import

import functools
import logging

from jaqpotpy.colorlog.colorlog import ColoredFormatter

BASIC_FORMAT = "%(log_color)s%(levelname)s%(reset)s:%(name)s:%(message)s"


def basicConfig(**kwargs):
    """Call ``logging.basicConfig`` and override the formatter it creates."""
    logging.basicConfig(**kwargs)
    logging._acquireLock()
    try:
        stream = logging.root.handlers[0]
        stream.setFormatter(
            ColoredFormatter(
                fmt=kwargs.get("format", BASIC_FORMAT),
                datefmt=kwargs.get("datefmt", None),
            )
        )
    finally:
        logging._releaseLock()


def ensure_configured(func):
    """Modify a function to call ``basicConfig`` first if no handlers exist."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if len(logging.root.handlers) == 0:
            basicConfig()
        return func(*args, **kwargs)

    return wrapper


root = logging.root
getLogger = logging.getLogger
debug = ensure_configured(logging.debug)
info = ensure_configured(logging.info)
warning = ensure_configured(logging.warning)
error = ensure_configured(logging.error)
critical = ensure_configured(logging.critical)
log = ensure_configured(logging.log)
exception = ensure_configured(logging.exception)

StreamHandler = logging.StreamHandler



import polling2
import os

from jaqpotpy.api.jaqpot_api_client_builder import JaqpotApiHttpClientBuilder
from jaqpotpy.api.model_to_b64encoding import file_to_b64encoding
from jaqpotpy.api.openapi import (
    ModelApi,
    DatasetApi,
    Dataset,
    DatasetType,
    DatasetCSV,
    Configuration,
)
from jaqpotpy.exceptions.exceptions import (
    JaqpotApiException,
    JaqpotPredictionTimeoutException,
    JaqpotPredictionFailureException,
)
from jaqpotpy.helpers.logging import init_logger
from jaqpotpy.models import Model
from jaqpotpy.helpers.url_utils import add_subdomain

QSARTOOLBOX_CALCULATOR_MODEL_ID = 6
QSARTOOLBOX_MODEL_MODEL_ID = 1837
QSAR_PROFILER_MODEL_ID = 1842


class JaqpotApiClient:
    """Client for interacting with the Jaqpot API.

    This client provides methods to interact with various endpoints of the Jaqpot API,
    including retrieving models and datasets, and making synchronous and asynchronous predictions.

    Attributes
    ----------
    base_url : str
        The base URL for the Jaqpot API.
    api_url : str
        The API URL for the Jaqpot API.
    http_client : object
        The HTTP client used to make requests to the Jaqpot API.
    log : object
        The logger object for logging messages.
    """

    def __init__(self, base_url=None, api_url=None, create_logs=False):
        """Initialize the JaqpotApiClient.

        Parameters
        ----------
        base_url : str, optional
            The base URL for the Jaqpot API. Defaults to "https://jaqpot.org".
        api_url : str, optional
            The API URL for the Jaqpot API. If not provided, it is constructed using the base URL.
        create_logs : bool, optional
            Whether to create logs. Defaults to False.
        """
        # logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)
        self.log = init_logger(
            __name__, testing_mode=False, output_log_file=create_logs
        )
        if base_url:
            self.base_url = base_url
        else:
            self.base_url = "https://jaqpot.org"
        self.api_url = api_url or add_subdomain(self.base_url, "api")
        jaqpot_api_key = os.getenv("JAQPOT_API_KEY")
        jaqpot_api_secret = os.getenv("JAQPOT_API_SECRET")
        self.http_client = (
            JaqpotApiHttpClientBuilder(host=self.api_url)
            .build_with_api_keys(jaqpot_api_key, jaqpot_api_secret)
            .build()
        )

    def get_model_by_id(self, model_id) -> Model:
        """Get a model from Jaqpot by its ID.

        Parameters
        ----------
        model_id : int
            The ID of the model on Jaqpot.

        Returns
        -------
        Model
            The model object retrieved from Jaqpot.

        Raises
        ------
        JaqpotApiException
            If the request fails, an exception is raised with the error message and status code.
        """
        model_api = ModelApi(self.http_client)
        response = model_api.get_model_by_id_with_http_info(id=model_id)
        if response.status_code < 300:
            return response.data.to_dict()
        raise JaqpotApiException(
            message=response.data.to_dict().message,
            status_code=response.status_code.value,
        )

    def get_model_summary(self, model_id):
        """Get a summary of a model from Jaqpot by its ID.

        Parameters
        ----------
        model_id : int
            The ID of the model on Jaqpot.

        Returns
        -------
        dict
            A dictionary containing the model summary.

        Raises
        ------
        JaqpotApiException
            If the request fails, an exception is raised with the error message and status code.
        """
        response = self.get_model_by_id(model_id)
        if response.status_code < 300:
            model = response.data
            model_summary = {
                "name": model.name,
                "modelId": model.id,
                "description": model.description,
                "type": model.type,
                "independentFeatures": [
                    feature.name for feature in model.independent_features
                ],
                "dependentFeatures": [
                    feature.name for feature in model.dependent_features
                ],
            }
            return model_summary
        raise JaqpotApiException(
            message=response.data.to_dict().message,
            status_code=response.status_code.value,
        )

    def get_shared_models(self, page=None, size=None, sort=None, organization_id=None):
        """Get shared models from Jaqpot.

        Parameters
        ----------
        page : int, optional
            Page number for pagination.
        size : int, optional
            Number of models per page.
        sort : str, optional
            Sort order for models.
        organization_id : int, optional
            Organization ID to filter models.

        Returns
        -------
        list
            A list of shared models.

        Raises
        ------
        JaqpotApiException
            If the request fails, an exception is raised with the error message and status code.
        """
        model_api = ModelApi(self.http_client)
        response = model_api.get_shared_models_with_http_info(
            page=page, size=size, sort=sort, organization_id=organization_id
        )
        if response.status_code < 300:
            return response
        raise JaqpotApiException(
            message=response.data.to_dict().message,
            status_code=response.status_code.value,
        )

    def get_dataset_by_id(self, dataset_id) -> Dataset:
        """Get a dataset from Jaqpot by its ID.

        Parameters
        ----------
        dataset_id : int
            The ID of the dataset on Jaqpot.

        Returns
        -------
        Dataset
            The dataset object retrieved from Jaqpot.

        Raises
        ------
        JaqpotApiException
            If the request fails, an exception is raised with the error message and status code.
        """
        dataset_api = DatasetApi(self.http_client)
        response = dataset_api.get_dataset_by_id_with_http_info(id=dataset_id)
        if response.status_code < 300:
            return response.data
        raise JaqpotApiException(
            message=response.data.to_dict().message,
            status_code=response.status_code.value,
        )

    def predict_sync(self, model_id, dataset):
        """Make a synchronous prediction with a model on Jaqpot.

        Parameters
        ----------
        model_id : int
            The ID of the model on Jaqpot.
        dataset : list or dict
            The dataset to predict.

        Returns
        -------
        dict
            The prediction result.

        Raises
        ------
        JaqpotApiException
            If the request fails, an exception is raised with the error message and status code.
        JaqpotPredictionFailureException
            If the prediction fails, an exception is raised with the failure reason.
        """
        dataset = Dataset(
            type=DatasetType.PREDICTION,
            entry_type="ARRAY",
            input=dataset,
        )

        model_api = ModelApi(self.http_client)
        response = model_api.predict_with_model_with_http_info(
            model_id=model_id, dataset=dataset
        )
        if response.status_code < 300:
            dataset = self._get_dataset_with_polling(response)
            if dataset.status == "SUCCESS":
                return dataset.result
            elif dataset.status == "FAILURE":
                raise JaqpotPredictionFailureException(dataset.failure_reason)
        raise JaqpotApiException(
            message=response.data.to_dict().message,
            status_code=response.status_code.value,
        )

    def predict_async(self, model_id, dataset):
        """Make an asynchronous prediction with a model on Jaqpot.

        Parameters
        ----------
        model_id : int
            The ID of the model on Jaqpot.
        dataset : list or dict
            The dataset to predict.

        Returns
        -------
        int
            The ID of the dataset containing the prediction results.

        Raises
        ------
        JaqpotApiException
            If the request fails, an exception is raised with the error message and status code.
        """
        dataset = Dataset(
            type=DatasetType.PREDICTION,
            entry_type="ARRAY",
            input=dataset,
        )

        model_api = ModelApi(self.http_client)
        response = model_api.predict_with_model_with_http_info(
            model_id=model_id, dataset=dataset
        )
        if response.status_code < 300:
            dataset_location = response.headers["Location"]
            dataset_id = int(dataset_location.split("/")[-1])
            return dataset_id
        raise JaqpotApiException(
            message=response.data.to_dict().message,
            status_code=response.status_code.value,
        )

    def predict_with_csv_sync(self, model_id, csv_path):
        """Make a synchronous prediction with a model on Jaqpot using a CSV file.

        Parameters
        ----------
        model_id : int
            The ID of the model on Jaqpot.
        csv_path : str
            The path to the CSV file.

        Returns
        -------
        dict
            The prediction result.

        Raises
        ------
        JaqpotApiException
            If the request fails, an exception is raised with the error message and status code.
        JaqpotPredictionFailureException
            If the prediction fails, an exception is raised with the failure reason.
        """
        b64_dataset_csv = file_to_b64encoding(csv_path)
        dataset_csv = DatasetCSV(
            type=DatasetType.PREDICTION, input_file=b64_dataset_csv
        )
        model_api = ModelApi(self.http_client)
        response = model_api.predict_with_model_csv_with_http_info(
            model_id=model_id, dataset_csv=dataset_csv
        )
        if response.status_code < 300:
            dataset = self._get_dataset_with_polling(response)
            if dataset.status == "SUCCESS":
                return dataset.result
            elif dataset.status == "FAILURE":
                raise JaqpotPredictionFailureException(message=dataset.failure_reason)
        raise JaqpotApiException(
            message=response.data.to_dict().message,
            status_code=response.status_code.value,
        )

    def _get_dataset_with_polling(self, response):
        """Retrieve a dataset by polling until it is ready or a timeout occurs.

        Parameters
        ----------
        response : requests.Response
            The HTTP response object containing the dataset location in the headers.

        Returns
        -------
        dict
            The dataset retrieved from the server.

        Raises
        ------
        JaqpotPredictionTimeoutException
            If polling times out while waiting for the dataset to be ready.
        """
        dataset_location = response.headers["Location"]
        dataset_id = int(dataset_location.split("/")[-1])
        try:
            polling2.poll(
                lambda: self.get_dataset_by_id(dataset_id).status
                in ["SUCCESS", "FAILURE"],
                step=3,
                timeout=10 * 60,
            )
        except polling2.TimeoutException:
            raise JaqpotPredictionTimeoutException(
                message="Polling timed out while waiting for prediction result."
            )
        dataset = self.get_dataset_by_id(dataset_id)
        return dataset

    def qsartoolbox_calculator_predict_sync(self, smiles, calculator_guid):
        """Synchronously predict using the QSAR Toolbox calculator.

        Parameters
        ----------
        smiles : str
            The SMILES string representing the chemical structure.
        calculator_guid : str
            The unique identifier for the QSAR Toolbox calculator.

        Returns
        -------
        dict
            The prediction result from the QSAR Toolbox calculator.
        """
        dataset = [{"smiles": smiles, "calculatorGuid": calculator_guid}]
        prediction = self.predict_sync(QSARTOOLBOX_CALCULATOR_MODEL_ID, dataset)
        return prediction

    def qsartoolbox_qsar_model_predict_sync(self, smiles, qsar_guid):
        """Synchronously predict QSAR model results using the QSAR Toolbox.

        Parameters
        ----------
        smiles : str
            The SMILES string representing the chemical structure.
        qsar_guid : str
            The unique identifier for the QSAR model.

        Returns
        -------
        dict
            The prediction results from the QSAR model.
        """
        dataset = [{"smiles": smiles, "qsarGuid": qsar_guid}]
        prediction = self.predict_sync(QSARTOOLBOX_MODEL_MODEL_ID, dataset)
        return prediction

    def qsartoolbox_profiler_predict_sync(self, smiles, profiler_guid):
        """Synchronously predict using the QSAR Toolbox profiler.

        Parameters
        ----------
        smiles : str
            The SMILES string representing the chemical structure.
        profiler_guid : str
            The unique identifier for the profiler.

        Returns
        -------
        dict
            The prediction result from the QSAR profiler model.
        """
        dataset = [{"smiles": smiles, "profilerGuid": profiler_guid}]
        prediction = self.predict_sync(QSAR_PROFILER_MODEL_ID, dataset)
        return prediction



from jaqpotpy.api.jaqpot_api_http_client import JaqpotApiHttpClient


class JaqpotApiHttpClientBuilder:
    """
    Builder class for creating a JaqpotApiHttpClient instance.
    """

    def __init__(self, host):
        """
        Initialize the builder with the host.

        Args:
            host (str): The host URL for the Jaqpot API.
        """
        self.http_client = JaqpotApiHttpClient(host=host)

    def build_with_access_token(self, access_token):
        """
        Set the access token for the HTTP client.

        Args:
            access_token (str): The access token for authentication.

        Returns:
            JaqpotApiHttpClientBuilder: The builder instance.
        """
        self.http_client.set_access_token(access_token)
        return self

    def build_with_api_keys(self, api_key, api_secret):
        if api_key is None or api_secret is None:
            raise ValueError("api_key and api_secret must be set")
        """
        Set the API keys for the HTTP client.

        Args:
            client_key (str): The client key for authentication.
            client_secret (str): The client secret for authentication.

        Returns:
            JaqpotApiHttpClientBuilder: The builder instance.
        """
        self.http_client.set_default_header("X-Api-Key", api_key)
        self.http_client.set_default_header("X-Api-Secret", api_secret)
        return self

    def build(self):
        """
        Build and return the JaqpotApiHttpClient instance.

        Returns:
            JaqpotApiHttpClient: The configured HTTP client instance.
        """
        return self.http_client



# coding: utf-8

"""
    Jaqpot API

    A modern RESTful API for model management and prediction services, built using Spring Boot and Kotlin. Supports seamless integration with machine learning workflows.

    The version of the OpenAPI document: 1.0.0
    Contact: upci.ntua@gmail.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from typing_extensions import Annotated
from jaqpotpy.api.openapi.models.feature_possible_value import FeaturePossibleValue
from jaqpotpy.api.openapi.models.feature_type import FeatureType
from typing import Optional, Set
from typing_extensions import Self

class PartiallyUpdateModelFeatureRequest(BaseModel):
    """
    PartiallyUpdateModelFeatureRequest
    """ # noqa: E501
    name: Annotated[str, Field(strict=True, max_length=255)] = Field(description="A name for the feature that will appear on top of the form field")
    units: Optional[StrictStr] = Field(default=None, description="The units that this feature is using")
    range: Optional[StrictStr] = Field(default=None, description="The range that this feature is using")
    description: Optional[Annotated[str, Field(strict=True, max_length=2000)]] = None
    feature_type: FeatureType = Field(alias="featureType")
    possible_values: Optional[Annotated[List[FeaturePossibleValue], Field(max_length=1000)]] = Field(default=None, alias="possibleValues")
    __properties: ClassVar[List[str]] = ["name", "units", "range", "description", "featureType", "possibleValues"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of PartiallyUpdateModelFeatureRequest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in possible_values (list)
        _items = []
        if self.possible_values:
            for _item_possible_values in self.possible_values:
                if _item_possible_values:
                    _items.append(_item_possible_values.to_dict())
            _dict['possibleValues'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of PartiallyUpdateModelFeatureRequest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "name": obj.get("name"),
            "units": obj.get("units"),
            "range": obj.get("range"),
            "description": obj.get("description"),
            "featureType": obj.get("featureType"),
            "possibleValues": [FeaturePossibleValue.from_dict(_item) for _item in obj["possibleValues"]] if obj.get("possibleValues") is not None else None
        })
        return _obj





"""Test basic molecular features."""

import unittest
import numpy as np
from jaqpotpy.descriptors.molecular import RDKitDescriptors
# pylint: disable=no-member


class TestRDKitDescriptors(unittest.TestCase):
    """Test RDKitDescriptors."""

    def setUp(self):
        """Set up tests."""
        from rdkit import Chem
        from rdkit.Chem import Descriptors

        smiles = "CC(=O)OC1=CC=CC=C1C(=O)O"
        self.mol = Chem.MolFromSmiles(smiles)
        self.featurizer = RDKitDescriptors()
        self.all_desc = Descriptors.descList
        self.first_desc = 10.6119

    # @unittest.skip("This test needs refactoring")
    def test_rdkit_descriptors(self):
        """Test simple descriptors."""

        featurizer = RDKitDescriptors()
        descriptors = featurizer([self.mol])
        assert descriptors.shape == (1, len(self.all_desc))
        assert np.allclose(descriptors[0][0], self.first_desc, atol=0.1)

    # # # @unittest.skip("This test needs refactoring")
    def test_rdkit_descriptors_on_smiles(self):
        """Test invocation on raw smiles."""
        featurizer = RDKitDescriptors()
        descriptors = featurizer("CC(=O)OC1=CC=CC=C1C(=O)O")
        assert descriptors.shape == (1, len(self.all_desc))
        assert np.allclose(descriptors[0][0], self.first_desc, atol=0.1)

    # @unittest.skip("This test needs refactoring")
    def test_rdkit_descriptors_on_smiles_df(self):
        """Test invocation on raw smiles."""
        featurizer = RDKitDescriptors()
        descriptors = featurizer.featurize_dataframe("CC(=O)OC1=CC=CC=C1C(=O)O")
        assert descriptors.shape == (1, 210)
        featurizer = RDKitDescriptors()
        descriptors = featurizer.featurize_dataframe(
            ["CC(=O)OC1=CC=CC=C1C(=O)O", "CC(=O)OC1=CC=CC=C1C(=O)O"]
        )
        assert descriptors.shape == (2, 210)

    # @unittest.skip("This test needs refactoring")
    def test_rdkit_descriptors_with_use_fragment(self):
        """Test with use_fragment"""

        featurizer = RDKitDescriptors(use_fragment=False)
        descriptors = featurizer(self.mol)
        # assert descriptors.shape == (1, len(featurizer.descriptors))
        assert len(descriptors) < len(self.all_desc)
        assert np.allclose(descriptors[0][0], self.first_desc, atol=0.1)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.insert(
    0, os.path.abspath("../..")
)  # Adds the parent directory of the docs folder

project = "jaqpotpy"
copyright = "2024, UPCI NTUA"
author = "UPCI NTUA"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
    "sphinx_markdown_builder",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "*test*"]

language = "en"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]

# -- Options for todo extension ----------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/todo.html#configuration

todo_include_todos = True



from __future__ import absolute_import

from jaqpotpy.colorlog.colorlog import (
    escape_codes,
    default_log_colors,
    ColoredFormatter,
    LevelFormatter,
    TTYColoredFormatter,
)

from jaqpotpy.colorlog.logging import (
    basicConfig,
    root,
    getLogger,
    log,
    debug,
    info,
    warning,
    error,
    exception,
    critical,
    StreamHandler,
)

__all__ = (
    "ColoredFormatter",
    "default_log_colors",
    "escape_codes",
    "basicConfig",
    "root",
    "getLogger",
    "debug",
    "info",
    "warning",
    "error",
    "exception",
    "critical",
    "log",
    "exception",
    "StreamHandler",
    "LevelFormatter",
    "TTYColoredFormatter",
)



# coding: utf-8

"""
    Jaqpot API

    A modern RESTful API for model management and prediction services, built using Spring Boot and Kotlin. Supports seamless integration with machine learning workflows.

    The version of the OpenAPI document: 1.0.0
    Contact: upci.ntua@gmail.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import json
from enum import Enum
from typing_extensions import Self


class ModelType(str, Enum):
    """
    ModelType
    """

    """
    allowed enum values
    """
    SKLEARN = 'SKLEARN'
    TORCH_ONNX = 'TORCH_ONNX'
    TORCHSCRIPT = 'TORCHSCRIPT'
    R_BNLEARN_DISCRETE = 'R_BNLEARN_DISCRETE'
    R_CARET = 'R_CARET'
    R_GBM = 'R_GBM'
    R_NAIVE_BAYES = 'R_NAIVE_BAYES'
    R_PBPK = 'R_PBPK'
    R_RF = 'R_RF'
    R_RPART = 'R_RPART'
    R_SVM = 'R_SVM'
    R_TREE_CLASS = 'R_TREE_CLASS'
    R_TREE_REGR = 'R_TREE_REGR'
    QSAR_TOOLBOX_CALCULATOR = 'QSAR_TOOLBOX_CALCULATOR'
    QSAR_TOOLBOX_QSAR_MODEL = 'QSAR_TOOLBOX_QSAR_MODEL'
    QSAR_TOOLBOX_PROFILER = 'QSAR_TOOLBOX_PROFILER'

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of ModelType from a JSON string"""
        return cls(json.loads(json_str))





# coding: utf-8

"""
    Jaqpot API

    A modern RESTful API for model management and prediction services, built using Spring Boot and Kotlin. Supports seamless integration with machine learning workflows.

    The version of the OpenAPI document: 1.0.0
    Contact: upci.ntua@gmail.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import json
from enum import Enum
from typing_extensions import Self


class ErrorCode(str, Enum):
    """
    ErrorCode
    """

    """
    allowed enum values
    """
    ENUM_1001 = '1001'

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of ErrorCode from a JSON string"""
        return cls(json.loads(json_str))


